#ifndef bempp_p0_discontinuous_shapeset_h
#define bempp_p0_discontinuous_shapeset_h

#include "bempp_base_types.h"

inline void p0_discontinuous_evaluate(const REALTYPE2* localPoint, REALTYPE* result)
{
    *result = M_ONE;

}


#endif