#ifndef bempp_spaces_h
#define bempp_spaces_h

#include "bempp_base_types.h"

#include "p0_discontinuous_shapeset.h"
#include "p1_discontinuous_shapeset.h"
#include "rwg0_shapeset.h"
#include "snc0_shapeset.h"

#define BASIS(name, modus) CAT(name, _ ## modus)


#endif