"""Define various test shapes."""

from .shapes import regular_sphere
from .shapes import cube
from .shapes import multitrace_cube
from .shapes import reference_triangle
from .shapes import sphere
from .shapes import ellipsoid
from .shapes import cuboid
from .shapes import almond
from .shapes import reentrant_cube
from .shapes import screen
from .shapes import cylinders
from .shapes import multitrace_ellipsoid
from .shapes import multitrace_sphere
