"""Definitions of finite element spaces.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/function_spaces.html
"""

from .space import function_space
