"""Functions for creating potential operators.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/potential_operators.html.
"""

from . import laplace
from . import helmholtz
from . import maxwell
from . import modified_helmholtz
