"""Functions for creating boundary integral operators.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/boundary_operators.html.
"""

from . import laplace
from . import helmholtz
from . import modified_helmholtz
from . import maxwell
from . import sparse
